from datetime import datetime
from typing import Any, Callable, ClassVar, Dict, Generic, Hashable, TypeVar, Union
from xml.etree.ElementTree import Element

from gpxpy.gpx import GPXTrackPoint
from pydantic import GetCoreSchemaHandler
from pydantic_core import CoreSchema, core_schema

from geo_track_analyzer.exceptions import GPXPointExtensionError

_KT = TypeVar("_KT", bound=Hashable)
_VT = TypeVar("_VT")


class BackFillExtensionDict(dict[_KT, list[_VT | None]], Generic[_KT, _VT]):
    def __missing__(self, key: _KT) -> list[_VT | None]:
        value: list[_VT | None] = []
        if self and key not in self:
            value = [None] * min(len(v) for v in self.values())
        self[key] = value
        return value

    def fill(self, data: dict[_KT, _VT | None]) -> None:
        for key, value in data.items():
            self[key].append(value)
        for key in [_key for _key in self if _key not in data]:
            self[key].append(None)


class ExtensionFieldElement(Element):
    def __init__(self, name: str, text: str | None) -> None:
        super().__init__(name)
        self.text = text


def get_extended_track_point(
    lat: float,
    lng: float,
    ele: None | float,
    timestamp: None | datetime,
    extensions: Dict[str, Union[str, float, int]],
) -> GPXTrackPoint:
    """
    Create a GPXTrackPoint with extended data fields.

    :param lat: Latitude of the track point.
    :param lng: Longitude of the track point.
    :param ele: Elevation of the track point (None if not available).
    :param timestamp: Timestamp of the track point (None if not available).
    :param extensions: Dictionary of extended data fields (key-value pairs).

    :return: GPXTrackPoint with specified attributes and extended data fields.
    """
    this_point = GPXTrackPoint(lat, lng, elevation=ele, time=timestamp)
    for key, value in extensions.items():
        this_point.extensions.append(
            ExtensionFieldElement(name=key, text=None if value is None else str(value))
        )

    return this_point


def get_extension_value(point: GPXTrackPoint, key: str) -> str:
    for ext in point.extensions:
        if ext.tag == key:
            if ext.text is None:
                raise GPXPointExtensionError(
                    "Key %s was not initilized with a value" % key
                )
            return ext.text  # type: ignore

    raise GPXPointExtensionError("Key %s could not be found" % key)


def get_extensions_in_points(points: list[GPXTrackPoint]) -> set[str]:
    found_extensions: set[str] = set()
    for point in points:
        found_extensions.update({str(ext.tag) for ext in point.extensions})
    return found_extensions


def _points_eq(p1: GPXTrackPoint, p2: GPXTrackPoint) -> bool:
    base_values = (
        (p1.latitude == p2.latitude)
        and (p1.longitude == p2.longitude)
        and (p1.elevation == p2.elevation)
        and (p1.time == p2.time)
    )
    if not base_values:
        return False

    if len(p1.extensions) == 0 and len(p1.extensions) == len(p2.extensions):
        return True

    if len(p1.extensions) != len(p2.extensions):
        return False

    d1 = {e.tag: e.text for e in p1.extensions if isinstance(e, ExtensionFieldElement)}
    d2 = {e.tag: e.text for e in p2.extensions if isinstance(e, ExtensionFieldElement)}

    return d1 == d2


class GPXTrackPointAfterValidator:
    valid_float_params: ClassVar[list[str]] = [
        "latitude",
        "longitude",
        "elevation",
        "horizontal_dilution",
        "vertical_dilution",
        "position_dilution",
        "speed",
        "heartrate",
        "cadence",
        "power",
    ]
    valid_str_params: ClassVar[list[str]] = [
        "symbol",
        "comment",
        "name",
    ]
    valid_dt_params: ClassVar[list[str]] = ["time"]

    extensions: ClassVar[list[str]] = [
        "heartrate",
        "cadence",
        "power",
    ]
    extensions_type_map: ClassVar[dict[str, Callable[[str], Any]]] = {
        "heartrate": float,
        "cadence": float,
        "power": float,
    }

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,
        handler: GetCoreSchemaHandler,
    ) -> CoreSchema:
        def validate_from_dict(value: dict[str, Any]) -> GPXTrackPoint:
            conv_dict: dict[str, Any] = {}
            for key, val in value.items():
                if key in cls.valid_float_params:
                    conv_dict[key] = float(val)
                elif key in cls.valid_str_params:
                    conv_dict[key] = str(val)
                elif key in cls.valid_dt_params:
                    try:
                        conv_dict[key] = datetime.fromisoformat(val)
                    except TypeError:
                        raise ValueError(
                            f"Value for {key} must valid isoformatted string"
                        )
                else:
                    raise ValueError(f"{key} is no valid parameter for GPXTrackPoint")

            pnt = GPXTrackPoint(
                **{k: v for k, v in conv_dict.items() if k not in cls.extensions}
            )
            for key, val in {
                k: v for k, v in conv_dict.items() if k in cls.extensions
            }.items():
                pnt.extensions.append(ExtensionFieldElement(name=key, text=str(val)))

            return pnt

        def dump_to_dict(
            point: GPXTrackPoint,
            supported_keys: list[str],
            extensions: list[str],
            extensions_type_map: dict[str, Callable[[str], Any]],
        ) -> dict[str, Any]:
            dump_dict = {}
            for key in supported_keys:
                if key in extensions:
                    continue
                val = getattr(point, key)
                if val:  # Ignore None and empty lists
                    dump_dict[key] = val

            if point.extensions:
                for ext in extensions:
                    try:
                        val = get_extension_value(point, ext)
                    except GPXPointExtensionError:
                        pass
                    else:
                        dump_dict[ext] = extensions_type_map[ext](val)

            return dump_dict

        from_dict_schema = core_schema.chain_schema(
            [
                core_schema.dict_schema(),
                core_schema.no_info_plain_validator_function(validate_from_dict),
            ]
        )

        keys_supported = (
            cls.valid_dt_params + cls.valid_str_params + cls.valid_float_params
        )

        return core_schema.json_or_python_schema(
            json_schema=from_dict_schema,
            python_schema=core_schema.union_schema(
                [core_schema.is_instance_schema(GPXTrackPoint), from_dict_schema]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda instance: dump_to_dict(
                    instance, keys_supported, cls.extensions, cls.extensions_type_map
                )
            ),
        )



"""Analyze geospacial data tracks"""

from .cli import extract_track, update_elevation
from .enhancer import (
    ElevationEnhancer,
    Enhancer,
    EnhancerType,
    OpenElevationEnhancer,
    OpenTopoElevationEnhancer,
    get_enhancer,
)
from .track import (
    ByteTrack,
    FITTrack,
    GeoJsonTrack,
    GPXFileTrack,
    PyTrack,
    SegmentTrack,
    Track,
)

__all__ = [
    "ByteTrack",
    "ElevationEnhancer",
    "Enhancer",
    "EnhancerType",
    "FITTrack",
    "GPXFileTrack",
    "GeoJsonTrack",
    "OpenElevationEnhancer",
    "OpenTopoElevationEnhancer",
    "PyTrack",
    "SegmentTrack",
    "Track",
    "extract_track",
    "get_enhancer",
    "update_elevation",
]

__version__ = "2.0.1"



import logging
from copy import copy
from datetime import timedelta
from typing import Any, Callable, Dict, Literal, Union

import numpy as np
import pandas as pd
from gpxpy.gpx import GPXTrack, GPXTrackPoint, GPXTrackSegment

from geo_track_analyzer.exceptions import GPXPointExtensionError
from geo_track_analyzer.model import Zones
from geo_track_analyzer.utils.internal import get_extension_value
from geo_track_analyzer.utils.model import format_zones_for_digitize

logger = logging.getLogger(__name__)


def _recalc_cumulated_columns(data: pd.DataFrame) -> pd.DataFrame:
    data = data.copy()
    data.cum_time = data.time.cumsum()
    data.cum_distance = data.distance.cumsum()

    cum_time_moving: list[None] | list[float] = []
    cum_time_stopped: list[None] | list[float] = []
    cum_distance_moving = []
    cum_distance_stopped = []
    for idx, rcrd in enumerate(data.to_dict("records")):
        if idx == 0:
            if rcrd["time"] is None:
                cum_time_moving.append(None)  # type: ignore
                cum_time_stopped.append(None)  # type: ignore
            else:
                if rcrd["moving"]:
                    cum_time_moving.append(rcrd["time"])
                    cum_time_stopped.append(0)
                else:
                    cum_time_moving.append(0)
                    cum_time_stopped.append(rcrd["time"])

            cum_distance_moving.append(rcrd["distance"] if rcrd["moving"] else 0)
            cum_distance_stopped.append(0 if rcrd["moving"] else rcrd["distance"])
        else:
            if rcrd["time"] is None:
                cum_time_moving.append(None)  # type: ignore
                cum_time_stopped.append(None)  # type: ignore
            else:
                if rcrd["moving"]:
                    cum_time_moving.append(cum_time_moving[-1] + rcrd["time"])
                    cum_time_stopped.append(cum_time_stopped[-1])
                else:
                    cum_time_moving.append(cum_time_moving[-1])
                    cum_time_stopped.append(cum_time_stopped[-1] + rcrd["time"])

            cum_distance_moving.append(
                cum_distance_moving[-1] + (rcrd["distance"] if rcrd["moving"] else 0)
            )
            cum_distance_stopped.append(
                cum_distance_stopped[-1] + (0 if rcrd["moving"] else rcrd["distance"])
            )

    data.cum_time_moving = cum_time_moving
    data.cum_time_stopped = cum_time_stopped
    data.cum_distance_moving = cum_distance_moving
    data.cum_distance_stopped = cum_distance_stopped

    return data


def get_processed_track_data(
    track: GPXTrack,
    stopped_speed_threshold: float = 1,
    connect_segments: Literal["full", "forward"] = "forward",
    heartrate_zones: None | Zones = None,
    power_zones: None | Zones = None,
    cadence_zones: None | Zones = None,
    extensions: set[str] | None = None,
    require_extensions: set[str] | None = None,
) -> tuple[float, float, float, float, pd.DataFrame]:
    """
    Process GPX track data and return a tuple of calculated values. The connect_segmen
    argument determines connection between segments in the DataFrame will be handled. As
    default each segment in the Track will be processed individually and the returned.
    Choosing forward will add the first point form the next segment to the end of each
    segment. So that each segment ends with the first point of the next segment.
    Choosingfull will also add the last point from the previous segment (which is the
    first point of the current segment) to the segment. This will result in a duplicated
    row in the DataFrame but can be usefull for plotting so that no gap is intrudoced by
    the conversion.

    :param track: GPXTrack to process
    :param stopped_speed_threshold: Threshold in km/h for speeds counting as stopped,
        default is 1 km/h
    :param connect_segments: Option to connect segments, choices are "full" or "forward"
        default is forward
    :return: Tuple containing track time, track distance, stopped time, stopped
        distance, and track data as a DataFrame
    :raises RuntimeError: If track has no segments
    """
    if connect_segments not in ["full", "forward"]:
        raise ValueError("connect_segments must be full or forward")
    track_time: float = 0
    track_distance: float = 0
    track_stopped_time: float = 0
    track_stopped_distance: float = 0
    track_data: None | pd.DataFrame = None

    for i_segment, segment in enumerate(track.segments):
        extend_segment_post = None
        extend_segment_start = None
        if connect_segments in ["full", "forward"]:
            try:
                next_segmnet = track.segments[i_segment + 1]
            except IndexError:
                pass
            else:
                extend_segment_post = [next_segmnet.points[0]]
        if i_segment > 0 and connect_segments in ["full"]:
            try:
                prev_segmnet = track.segments[i_segment - 1]
            except IndexError:
                pass
            else:
                extend_segment_start = [prev_segmnet.points[-1]]
        (
            time,
            distance,
            stopped_time,
            stopped_distance,
            _data,
        ) = get_processed_segment_data(
            segment,
            stopped_speed_threshold,
            extend_segment_start=extend_segment_start,
            extend_segment_end=extend_segment_post,
            extensions=extensions,
            require_extensions=require_extensions,
        )

        track_time += time
        track_distance += distance
        track_stopped_time += stopped_time
        track_stopped_distance += stopped_distance

        data = _data.copy()
        data["segment"] = i_segment

        if track_data is None:
            track_data = data
        else:
            track_data = pd.concat([track_data, data]).reset_index(drop=True)

    # Not really possible but keeps linters happy
    if track_data is None:
        raise RuntimeError("Track has no segments")

    # Recalculate all cumulated columns over the segments
    track_data = _recalc_cumulated_columns(track_data)

    if heartrate_zones is not None:
        track_data = add_zones_to_dataframe(track_data, "heartrate", heartrate_zones)
    if power_zones is not None:
        track_data = add_zones_to_dataframe(track_data, "power", power_zones)
    if cadence_zones is not None:
        track_data = add_zones_to_dataframe(track_data, "cadence", cadence_zones)

    return (
        track_time,
        track_distance,
        track_stopped_time,
        track_stopped_distance,
        track_data,
    )


def get_processed_segment_data(
    segment: GPXTrackSegment,
    stopped_speed_threshold: float = 1,
    extend_segment_start: None | list[GPXTrackPoint] = None,
    extend_segment_end: None | list[GPXTrackPoint] = None,
    heartrate_zones: None | Zones = None,
    power_zones: None | Zones = None,
    cadence_zones: None | Zones = None,
    extensions: set[str] | None = None,
    require_extensions: set[str] | None = None,
) -> tuple[float, float, float, float, pd.DataFrame]:
    """
    Calculate the speed and distance from point to point for a segment. This follows
    the implementation of the get_moving_data method in the implementation of
    gpx.GPXTrackSegment

    :param segment: GPXTrackSegment to process
    :param stopped_speed_threshold: Threshold in km/h for speeds counting as moving,
        default is 1 km/h
    :param extend_segment_start: Additional points to add at the start of the segment
    :param extend_segment_end: Additional points to add at the end of the segment

    :return: Tuple containing segment time, segment distance, stopped time, stopped
        distance, and segment data as a DataFrame
    """
    if extend_segment_start or extend_segment_end:
        segment = segment.clone()

    if extend_segment_start:
        extend_segment_start.extend(segment.points)
        segment.points = extend_segment_start
    if extend_segment_end:
        segment.points.extend(extend_segment_end)

    threshold_ms = stopped_speed_threshold / 3.6

    data: Dict[str, list[None | Union[float, bool]]] = {
        "latitude": [],
        "longitude": [],
        "elevation": [],
        "speed": [],
        "distance": [],
        "time": [],
        "cum_time": [],
        "cum_time_moving": [],
        "cum_time_stopped": [],
        "cum_distance": [],
        "cum_distance_moving": [],
        "cum_distance_stopped": [],
        "moving": [],
    }

    _extensions: set[str] = set() if extensions is None else extensions
    _require_extensions: set[str] = (
        set() if require_extensions is None else require_extensions
    )

    data_extensions = copy(_extensions)
    data_extensions.update(_require_extensions)
    if data_extensions:
        for key in data_extensions:
            data[key] = []

    if segment.has_times():
        (
            time,
            distance,
            stopped_time,
            stopped_distance,
            data,
        ) = _get_processed_data_w_time(
            segment,
            data,
            threshold_ms,
            extensions=data_extensions,
        )
    else:
        distance, data = _get_processed_data_wo_time(
            segment,
            data,
            extensions=data_extensions,
        )
        time, stopped_distance, stopped_time = 0, 0, 0

    data_df = pd.DataFrame(data)

    if heartrate_zones is not None:
        data_df = add_zones_to_dataframe(data_df, "heartrate", heartrate_zones)
    if power_zones is not None:
        data_df = add_zones_to_dataframe(data_df, "power", power_zones)
    if cadence_zones is not None:
        data_df = add_zones_to_dataframe(data_df, "cadence", cadence_zones)

    return (time, distance, stopped_time, stopped_distance, data_df)


def _get_processed_data_w_time(
    segment: GPXTrackSegment,
    data: Dict[str, list[Any]],
    threshold_ms: float,
    extensions: set[str],
) -> tuple[float, float, float, float, Dict[str, list[Any]]]:
    time = 0.0
    stopped_time = 0.0

    distance = 0.0
    stopped_distance = 0.0

    cum_time = 0
    cum_time_moving = 0
    cum_time_stopped = 0

    cum_distance = 0
    cum_moving = 0
    cum_stopped = 0
    for previous, point in zip(segment.points, segment.points[1:]):
        # Ignore first and last point
        if point.time and previous.time:
            timedelta = point.time - previous.time

            if point.elevation and previous.elevation:
                point_distance = point.distance_3d(previous)
            else:
                point_distance = point.distance_2d(previous)

            seconds = timedelta.total_seconds()
            if seconds > 0 and point_distance is not None:
                if point_distance:
                    is_stopped = (point_distance / seconds) <= threshold_ms
                else:
                    is_stopped = True

                data["distance"].append(point_distance)

                if is_stopped:
                    stopped_time += seconds
                    stopped_distance += point_distance
                    cum_stopped += point_distance
                    data["moving"].append(False)
                else:
                    time += seconds
                    distance += point_distance
                    cum_moving += point_distance
                    data["moving"].append(True)

                cum_distance += point_distance
                cum_time += seconds
                data["time"].append(seconds)
                data["cum_time"].append(cum_time)

                data["cum_distance"].append(cum_distance)
                data["cum_distance_moving"].append(cum_moving)
                data["cum_distance_stopped"].append(cum_stopped)

                data["latitude"].append(point.latitude)
                data["longitude"].append(point.longitude)
                if point.has_elevation():
                    data["elevation"].append(point.elevation)
                else:
                    data["elevation"].append(None)

                if not is_stopped:
                    data["speed"].append(point_distance / seconds)
                    cum_time_moving += seconds
                    # data["cum_time_moving"].append(cum_time_moving)
                    # data["cum_time_stopped"].append(None)
                else:
                    data["speed"].append(None)
                    # data["cum_time_moving"].append(None)
                    cum_time_stopped += seconds
                    # data["cum_time_stopped"].append(cum_time_stopped)

                data["cum_time_moving"].append(cum_time_moving)
                data["cum_time_stopped"].append(cum_time_stopped)

                for key in extensions:
                    try:
                        data[key].append(float(get_extension_value(point, key)))
                    except GPXPointExtensionError:
                        data[key].append(None)

    # print(extensions)
    # for key, values in data.items():
    #     print(key, len(values))
    return time, distance, stopped_time, stopped_distance, data


def _get_processed_data_wo_time(
    segment: GPXTrackSegment,
    data: Dict[str, list[Any]],
    extensions: set[str],
) -> tuple[float, Dict[str, list[Any]]]:
    cum_distance = 0
    distance = 0.0
    for previous, point in zip(segment.points, segment.points[1:]):
        if point.elevation and previous.elevation:
            point_distance = point.distance_3d(previous)
        else:
            point_distance = point.distance_2d(previous)
        if point_distance is not None:
            distance += point_distance

            data["distance"].append(point_distance)
            data["latitude"].append(point.latitude)
            data["longitude"].append(point.longitude)
            if point.has_elevation():
                data["elevation"].append(point.elevation)
            else:
                data["elevation"].append(None)

            cum_distance += point_distance
            data["time"].append(None)
            data["cum_time"].append(None)
            data["cum_time_moving"].append(None)
            data["cum_time_stopped"].append(None)
            data["cum_distance"].append(cum_distance)
            data["cum_distance_moving"].append(cum_distance)
            data["cum_distance_stopped"].append(None)
            data["speed"].append(None)
            data["moving"].append(True)

            for key in extensions:
                try:
                    data[key].append(float(get_extension_value(point, key)))
                except GPXPointExtensionError:
                    data[key].append(None)

    return distance, data


def split_data_by_time(
    data: pd.DataFrame,
    split_at: timedelta,
    moving_only: bool = True,
    method: Literal["first", "closest", "interploation"] = "closest",
) -> pd.DataFrame:
    return split_data(
        data=data,
        split_by="time",
        split_at=split_at.total_seconds(),
        moving_only=moving_only,
        method=method,
    )


def split_data(
    data: pd.DataFrame,
    split_by: Literal["distance", "time"],
    split_at: float,
    moving_only: bool = True,
    method: Literal["first", "closest", "interploation"] = "closest",
) -> pd.DataFrame:
    split_idx_finder: Callable[[pd.Series, float], int]
    if method == "closest":
        split_idx_finder = lambda s, v: np.abs(s.to_numpy() - v).argmin()  # type: ignore
    # TODO: Implement method in which the plotting point is interpolat to the ecxat val
    elif method == "interploation":
        raise NotImplementedError("Interploation splitting method not implemented")
    else:
        split_idx_finder = lambda s, v: s[s < v].index.max()

    data = data.copy()

    if split_by == "distance":
        column = "cum_distance_moving" if moving_only else "cum_distance"
    else:
        column = "cum_time_moving" if moving_only else "cum_time"

    max_value = data[column].max()

    if split_at > max_value:
        logger.warning(
            "Data can not be split further by %s with passed value %s",
            split_by,
            split_at,
        )
        return data

    split_vals = [split_at]
    while split_vals[-1] < max_value:
        split_vals.append(split_vals[-1] + split_at)

    split_vals = split_vals[:-1]

    logger.debug("Splitting into %s segments", len(split_vals))

    last_split = 0
    i_segement = 0
    for val in split_vals:
        split_idx = split_idx_finder(data[column], val)
        data.loc[last_split:split_idx, "segment"] = i_segement
        last_split = split_idx + 1
        i_segement += 1

    if last_split != data.index.max() + 1:
        data.loc[last_split : data.index.max() + 1, "segment"] = i_segement

    return data


def add_zones_to_dataframe(
    data: pd.DataFrame, metric: Literal["heartrate", "power", "cadence"], zones: Zones
) -> pd.DataFrame:
    zone_bins, names, zone_colors = format_zones_for_digitize(zones)
    if zone_colors is None:
        try:
            import plotly
        except ModuleNotFoundError:
            if len(names) > 8:
                raise RuntimeError("Out of automatic colors for more than 8 zones")
            zone_colors = [
                "#FF0000",
                "#F71D1D",
                "#EF3A3A",
                "#E75656",
                "#DF7373",
                "#D78F8F",
                "#CFABAB",
                "#C7C7C7",
            ][0 : len(names)]
        else:
            zone_colors = plotly.colors.sample_colorscale("viridis", len(names))

    metric_data = data[metric][~data[metric].isna()]
    binned_metric = pd.Series(
        np.digitize(metric_data, zone_bins), index=metric_data.index
    )
    data[f"{metric}_zones"] = binned_metric.apply(lambda v: names[v - 1])
    data[f"{metric}_zone_colors"] = binned_metric.apply(lambda v: zone_colors[v - 1])
    data.loc[data[data[metric].isna()].index, f"{metric}_zones"] = names[0]
    data.loc[data[data[metric].isna()].index, f"{metric}_zone_colors"] = zone_colors[0]

    return data



import logging
from collections import defaultdict
from datetime import datetime
from functools import lru_cache
from typing import Any, Generator

from sqlalchemy import Engine, text
from sqlalchemy.engine import Connection

from geo_track_analyzer.exceptions import (
    DBTrackInitializationError,
    GPXPointExtensionError,
)
from geo_track_analyzer.track import PyTrack, Track
from geo_track_analyzer.utils.internal import get_extension_value

logger = logging.getLogger(__name__)


def create_tables(
    engine: Engine,
    schema: str,
    track_table: str,
    points_table: str,
    extensions: list[tuple[str, str]] | None = None,
) -> None:
    """
    Create the tables for the tracks and points using the passed Engine

    :param engine: SQLLalchemy Engine for the Postgres database
    :param schema: Name of the schema in the database
    :param track_table: Name of the table containing track information
    :param points_table: Name of the table containing the points of
        all tracks
    :param extensions: Extensions columns (tuple of name and type). If None
        is passed heartrate, cadence, power, and temperature are created
    """
    track_stmt = f"""
        CREATE TABLE IF NOT EXISTS {schema}.{track_table} (
            id SERIAL PRIMARY KEY,
            track_name TEXT,
            track_date DATE,
            source TEXT
        );
    """

    if extensions is None:
        extensions = [
            ("heartrate", "INTEGER"),
            ("cadence", "INTEGER"),
            ("power", "INTEGER"),
            ("temperature", "DOUBLE PRECISION"),
        ]

    points_stmt = f"""
        CREATE TABLE IF NOT EXISTS {schema}.{points_table} (
        id SERIAL PRIMARY KEY,
        track_id INTEGER REFERENCES {schema}.{track_table}(id),
        segment_id INTEGER NOT NULL,
        geom GEOGRAPHY(POINT, 4326),
        elevation DOUBLE PRECISION DEFAULT NULL,
        time TIMESTAMP WITH TIME ZONE,
        {",".join([f"{n} {t} DEFAULT NULL" for n, t in extensions])},
        -- Add other extension fields as needed
        CONSTRAINT fk_track FOREIGN KEY (track_id) REFERENCES {schema}.{track_table}(id)
    );
    """

    index_stmt = f"""
    CREATE INDEX {points_table}_geom_idx ON {schema}.{points_table} USING GIST (geom);
    """

    with engine.connect() as conn:
        logger.debug("Table %s.%s created", schema, track_table)
        conn.execute(text(track_stmt))
        conn.commit()
        logger.debug("Table %s.%s created", schema, points_table)
        conn.execute(text(points_stmt))
        conn.commit()
        logger.debug("Index created")
        conn.execute(text(index_stmt))
        conn.commit()


def get_track_data(
    track: Track, track_id: int, batch_size: int = 100
) -> Generator[
    list[dict[str, Any]],
    None,
    None,
]:
    """
    Generator function that yields batches of track data tuples.

    Args:
        track: Track object containing GPS data
        batch_size: Number of data points to include in each batch

    Yields:
        List of tuples containing track data in batches
    """
    current_batch = []
    for i, segment in enumerate(track.track.segments):
        for point in segment.points:
            long, lat, ele, time = (
                point.longitude,
                point.latitude,
                point.elevation,
                point.time,
            )
            ext = {}
            for key in ["heartrate", "cadence", "power", "temperature"]:
                try:
                    ext[key] = float(get_extension_value(point, key))
                except GPXPointExtensionError:
                    ext[key] = None
            _data = {
                "track_id": track_id,
                "segment_id": i,
                "lat": lat,
                "long": long,
                "time": time,
                "ele": ele,
            }
            _data.update(ext)

            current_batch.append(_data)
            if len(current_batch) >= batch_size:
                yield current_batch
                current_batch = []

    # Yield any remaining data points in the final batch
    if current_batch:
        yield current_batch


@lru_cache(maxsize=5)
def _find_extensions(conn: Connection, schema: str, table: str) -> list[str]:
    base_cols = [
        "id",
        "track_id",
        "segment_id",
        "geom",
        "elevation",
        "time",
    ]
    _columns = conn.execute(
        text(f"""
        SELECT column_name
        FROM information_schema.columns
        WHERE table_schema = '{schema}' AND table_name = '{table}'
        """)
    ).fetchall()

    extensions = [r[0] for r in _columns if r[0] not in base_cols]
    logger.debug("Found extenstions %s in %s.%s", extensions, schema, table)
    return extensions


def insert_track(
    track: Track,
    engine: Engine,
    schema: str,
    track_table: str,
    points_table: str,
    source: str,
) -> int | None:
    """
    Insert a track into the database

    :param track: Track to be inserted
    :param engine: SQLLalchemy Engine for the Postgres database
    :param schema: Name of the schema in the database
    :param track_table: Name of the table containing track information
    :param points_table: Name of the table containing the points of
        all tracks
    :param source: Set a source of the track in the track table.

    :return: None if insertion failed and track_id if successfull
    """
    start_time = track.track.get_time_bounds().start_time

    with engine.connect() as conn:
        track_stmt = f"""
            INSERT INTO {schema}.{track_table} (track_name, track_date, source)
            VALUES (:name, :date, :source) RETURNING id
        """
        new_track_id = conn.execute(
            text(track_stmt),
            [
                {
                    "name": track.track.name,
                    "date": None if start_time is None else start_time.date(),
                    "source": source,
                }
            ],
        ).fetchone()
        conn.commit()
        logger.info("Track entry created")
        if new_track_id is None:
            logger.error("Track insertion uncessefull")
            return None
        new_track_id = new_track_id[0]
        logger.debug("Inserted track with id %s", new_track_id)
        extensions = _find_extensions(conn, schema, points_table)
        points_stmt = f"""
            INSERT INTO {schema}.{points_table}
            (track_id, segment_id, geom,
             elevation, time, {",".join(extensions)})
            VALUES (:track_id, :segment_id,
                    ST_SetSRID(ST_MakePoint(:long, :lat), 4326)::geography,
                    :ele, :time,{",".join(map(lambda v: f":{v}", extensions))})
        """
        for point_batch in get_track_data(track, new_track_id, 100):
            conn.execute(text(points_stmt), point_batch)
            conn.commit()
        logger.info("All points inserted")
        return new_track_id


def _load_points(
    conn: Connection, track_id: int, schema: str, table: str, extensions: list[str]
) -> dict[int, dict[str, list[str] | list[float | None] | list[datetime | None]]]:
    ret_data = {}
    segment_stmt = (
        f"SELECT DISTINCT segment_id FROM {schema}.{table} WHERE track_id = {track_id}"
    )
    segment_ids = list([r[0] for r in conn.execute(text(segment_stmt)).fetchall()])
    if not segment_ids:
        raise DBTrackInitializationError(
            "Could not load any segments for track id %s" % track_id
        )
    for segment_id in segment_ids:
        stmt = f"""
            SELECT
                ST_Y(geom::geometry) AS latitude,
                ST_X(geom::geometry) AS longitude,
                elevation,
                time,
                {",".join(extensions)}
            FROM {schema}.{table}
            WHERE track_id = {track_id} AND
                  segment_id = {segment_id}
        """
        data = conn.execute(text(stmt)).fetchall()
        if not data:
            raise DBTrackInitializationError(
                "Could not load points for segment %s in track %s"
                % (segment_id, track_id)
            )
        segment_data = defaultdict(list)
        for d in data:
            for key, value in d._mapping.items():
                segment_data[key].append(value)
        ret_data[segment_id] = segment_data

    return ret_data


def load_track(
    track_id: int,
    engine: Engine,
    schema: str,
    track_table: str,
    points_table: str,
    **track_kwargs,
) -> Track:
    """
    Insert a track into the database

    :param track_id: id of the track in the database
    :param engine: SQLLalchemy Engine for the Postgres database
    :param schema: Name of the schema in the database
    :param track_table: Name of the table containing track information
    :param points_table: Name of the table containing the points of
        all tracks
    :param track_kwargs: Additional keyword arguments passed to the Track.

    :return: Track Object.
    """
    track_stmt = f"SELECT * FROM {schema}.{track_table} WHERE id = {track_id}"
    track = None
    with engine.connect() as conn:
        logger.debug("Loading track: %s from %s.%s", track_id, schema, track_table)
        _track_data = conn.execute(text(track_stmt)).fetchone()
        if _track_data is None:
            raise DBTrackInitializationError(
                "No track with id %s in %s.%s" % (track_id, schema, track_table)
            )
        track_data = _track_data._mapping

        logger.debug("Loading points from %s.%s", schema, points_table)
        extensions = _find_extensions(conn, schema, points_table)
        segments = _load_points(conn, track_id, schema, points_table, extensions)
        for i, segment_id in enumerate(sorted(segments.keys())):
            segment_data = segments[segment_id]
            extension_data = {}
            for extension in extensions:
                extension_data[extension] = segment_data[extension]
            _data = dict(
                points=list(zip(segment_data["latitude"], segment_data["longitude"])),
                elevations=segment_data["elevation"],
                times=segment_data["time"],
                extensions=extension_data,
            )
            if i == 0:
                track = PyTrack(
                    **_data,
                    **track_kwargs,
                )
            else:
                assert track is not None
                track.add_segmeent(**_data)  # type: ignore
    assert track is not None
    track.track.name = track_data["track_name"]
    return track



from itertools import pairwise
from typing import Literal

import pandas as pd

from geo_track_analyzer.exceptions import VisualizationSetupError
from geo_track_analyzer.processing import (
    get_processed_segment_data,
    get_processed_track_data,
)
from geo_track_analyzer.track import Track, logger


def extract_track_data_for_plot(
    track: Track,
    kind: str,
    require_elevation: list[str],
    intervals: None | int = None,
    connect_segments: Literal["full", "forward"] = "forward",
    extensions: set[str] | None = None,
    require_extensions: set[str] | None = None,
) -> pd.DataFrame:
    """Extract the data from a Track as DataFrame for plotting.

    :param track: Track object
    :param kind: Kind of plot
    :param require_elevation: List of kinds of plots that require elevation data to be
        present in the Track
    :param intervals: Optionally reduce the pp-distance in the track, defaults to None

    :return: DataFrame
    """
    if kind in require_elevation and not track.track.has_elevations():
        raise VisualizationSetupError(f"Track has so elevation so {kind} is not valid")
    _track = track.track

    if intervals is not None:
        if track.get_avg_pp_distance() >= intervals:
            logger.debug("Average pp distance larget than interval. Skipping reduction")
        else:
            _track = _track.clone()
            _track.reduce_points(intervals)

    _, _, _, _, data = get_processed_track_data(
        _track,
        connect_segments=connect_segments,
        heartrate_zones=track.heartrate_zones,
        power_zones=track.power_zones,
        cadence_zones=track.cadence_zones,
        extensions=extensions,
        require_extensions=require_extensions,
    )

    return data


def extract_multiple_segment_data_for_plot(
    track: Track,
    segments: list[int],
    kind: str,
    require_elevation: list[str],
    intervals: None | int = None,
    connect_segments: Literal["full", "forward"] = "forward",
    extensions: set[str] | None = None,
    require_extensions: set[str] | None = None,
) -> pd.DataFrame:
    """Extract the data for a two or more segments from a Track as DataFrame for
    plotting.

    :param track: Track object
    :param segments: Indices of the segments to be extracted
    :param kind: Kind of plot
    :param require_elevation: List of kinds of plots that require elevation data to be
        present in the Track
    :param intervals: Optionally reduce the pp-distance in the track, defaults to None

    :return: DataFrame
    """
    if len(segments) < 2:
        raise VisualizationSetupError("Pass at least two segment ids")
    if max(segments) >= track.n_segments or min(segments) < 0:
        raise VisualizationSetupError(
            f"Passed ids must be between 0 and {len(segments) - 1}. Got {segments}"
        )

    data = extract_track_data_for_plot(
        track=track,
        kind=kind,
        require_elevation=require_elevation,
        intervals=intervals,
        connect_segments=connect_segments,
        extensions=extensions,
        require_extensions=require_extensions,
    )

    return data[data.segment.isin(segments)]


def extract_segment_data_for_plot(
    track: Track,
    segment: int,
    kind: str,
    require_elevation: list[str],
    intervals: None | int = None,
    extensions: set[str] | None = None,
    require_extensions: set[str] | None = None,
) -> pd.DataFrame:
    """Extract the data for a segment from a Track as DataFrame for plotting.

    :param track: Track object
    :param segment: Index of the segment to be extracted
    :param kind: Kind of plot
    :param require_elevation: List of kinds of plots that require elevation data to be
        present in the Track
    :param intervals: Optionally reduce the pp-distance in the track, defaults to None

    :return: DataFrame
    """
    if kind in require_elevation and not track.track.segments[segment].has_elevations():
        raise VisualizationSetupError(
            f"Segment has so elevation so {kind} is not valid"
        )
    if kind == "map-segments":
        raise VisualizationSetupError("map-segments can only be done for full tracks")

    segement = track.track.segments[segment]

    if intervals is not None:
        if track.get_avg_pp_distance_in_segment(segment) >= intervals:
            logger.debug("Average pp distance larget than interval. Skipping reduction")
        else:
            segement = track.track.segments[segment].clone()
            segement.reduce_points(intervals)

    _, _, _, _, data = get_processed_segment_data(
        segement,
        heartrate_zones=track.heartrate_zones,
        power_zones=track.power_zones,
        cadence_zones=track.cadence_zones,
        extensions=extensions,
        require_extensions=require_extensions,
    )

    return data


def generate_distance_segments(data: pd.DataFrame, distance: float) -> pd.DataFrame:
    """Generate segments with the distance specified with the passed parameter. Segments
    present in the passed data will be replaced. Splitting is done based on the
    distance_moving value in the data. Segements are split closest to the passed
    distance. So extact cummulated distance in a segment depends on the pp-distance in
    the track. The last segmeent may be shorter than the passed distance.

    :param data: Dataframe create with extract_track_data_for_plot,
        extract_segment_data_for_plot, or extract_multiple_segment_data_for_plot
        methods.
    :param distance: Intended segmeent distance

    :return: Dataframe with updated segments
    """
    key = "cum_distance_moving"
    max_distance = data.iloc[-1][key]

    if max_distance < distance:
        data["segment"] = 0
        return data

    distances, segments_ids = [0.0], [0]
    while distances[-1] < max_distance:
        distances.append(distances[-1] + distance)
        segments_ids.append(segments_ids[-1] + 1)

    new_segments = data.segment.copy()
    new_segments.loc[:] = max(segments_ids)
    for (start, end), idx in zip(pairwise(distances), segments_ids):
        new_segments.iloc[data[(data[key] >= start) & (data[key] < end)].index] = idx

    data["segment"] = new_segments
    return data



try:
    import plotly  # noqa: F401
except ModuleNotFoundError:

    def raise_err_func(*args, **kwargs) -> None:
        raise RuntimeError("Install the visualization extra to use this functionality")

    plot_metrics = lambda *args, **kwargs: raise_err_func(*args, **kwargs)
    plot_segment_box_summary = lambda *args, **kwargs: raise_err_func(*args, **kwargs)
    plot_segment_summary = lambda *args, **kwargs: raise_err_func(*args, **kwargs)
    plot_segment_zones = lambda *args, **kwargs: raise_err_func(*args, **kwargs)
    plot_segments_on_map = lambda *args, **kwargs: raise_err_func(*args, **kwargs)
    plot_track_2d = lambda *args, **kwargs: raise_err_func(*args, **kwargs)
    plot_track_3d = lambda *args, **kwargs: raise_err_func(*args, **kwargs)
    plot_track_enriched_on_map = lambda *args, **kwargs: raise_err_func(*args, **kwargs)
    plot_track_line_on_map = lambda *args, **kwargs: raise_err_func(*args, **kwargs)
    plot_track_with_slope = lambda *args, **kwargs: raise_err_func(*args, **kwargs)
    plot_track_zones = lambda *args, **kwargs: raise_err_func(*args, **kwargs)
    plot_tracks_on_map = lambda *args, **kwargs: raise_err_func(*args, **kwargs)
else:
    from .interactive import plot_track_3d
    from .map import (
        plot_segments_on_map,
        plot_track_enriched_on_map,
        plot_track_line_on_map,
        plot_tracks_on_map,
    )
    from .metrics import plot_metrics
    from .profiles import plot_track_2d, plot_track_with_slope
    from .summary import (
        plot_segment_box_summary,
        plot_segment_summary,
        plot_segment_zones,
        plot_track_zones,
    )

__all__ = [
    "plot_metrics",
    "plot_segment_box_summary",
    "plot_segment_summary",
    "plot_segment_zones",
    "plot_segments_on_map",
    "plot_track_2d",
    "plot_track_3d",
    "plot_track_enriched_on_map",
    "plot_track_line_on_map",
    "plot_track_with_slope",
    "plot_track_zones",
    "plot_tracks_on_map",
]



import numpy as np
import numpy.typing as npt

from geo_track_analyzer.model import Zones


def format_zones_for_digitize(
    zones: Zones,
) -> tuple[npt.NDArray, list[str], None | list[str]]:
    vals = [-np.inf]
    if zones.intervals[0].name is None:
        names = [f"Zone 1 [0, {zones.intervals[0].end}]"]
    else:
        names = [f"{zones.intervals[0].name} [0, {zones.intervals[0].end}]"]
    if zones.intervals[0].color is None:
        colors = []
    else:
        colors = [zones.intervals[0].color]
    for i, interval in enumerate(zones.intervals[1:]):
        vals.append(float(interval.start))  # type: ignore
        base_name = f"Zone {i+2}" if interval.name is None else interval.name
        if interval.end is None:
            interval_str = f"[{interval.start}, \u221e]"
        else:
            interval_str = f"[{interval.start}, {interval.end}]"
        names.append(f"{base_name} {interval_str}")
        if interval.color is not None:
            colors.append(interval.color)
    vals.append(np.inf)

    _colors = None
    if len(colors) == len(names):
        _colors = colors

    return np.array(vals), names, _colors
